@@ -6,7 +6,7 @@
  *  used by the graphics utility of the FreeType test suite.
  *
  *  Written by Antoine Leca.
- *  Copyright (C) 1999-2020 by
+ *  Copyright (C) 1999-2023 by
  *  Antoine Leca, David Turner, Robert Wilhelm, and Werner Lemberg.
  *
  *  Borrowing liberally from the other FreeType drivers.
@@ -19,34 +19,27 @@
  *
  ******************************************************************/
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
+#define WIN32_LEAN_AND_MEAN
+#include <windows.h>
+
+#include "grobjs.h"
+#include "grdevice.h"
 
 /* define to activate OLPC swizzle */
 #define xxSWIZZLE
 
-#define WIN32_LEAN_AND_MEAN
-#include <windows.h>
-
-#include <grobjs.h>
-#include <grdevice.h>
 #ifdef SWIZZLE
-#include <grswizzle.h>
+# include "grswizzle.h"
 #endif
 
 /* logging facility */
-#include <stdarg.h>
-
-#define  DEBUGxxx
+#define  xxDEBUG
 
 #ifdef DEBUG
-#define LOG(x)  LogMessage##x
-#else
-#define LOG(x)  /* rien */
-#endif
+# include <stdio.h>
+# include <stdarg.h>
+# include <ctype.h>
 
-#ifdef DEBUG
   static void  LogMessage( const char*  fmt, ... )
   {
     va_list  ap;
@@ -55,14 +48,17 @@
     vfprintf( stderr, fmt, ap );
     va_end( ap );
   }
+
+#define LOG(x)  LogMessage##x
+#else
+#define LOG(x)  /* rien */
 #endif
-/*-------------------*/
 
 /*  Custom messages. */
+#define WM_STATUS  WM_USER+512
 #define WM_RESIZE  WM_USER+517
+#define WM_GR_KEY  WM_USER+519
 
-/* These values can be changed, but WIN_WIDTH should remain for now a  */
-/* multiple of 32 to avoid padding issues.                             */
 
   typedef struct  Translator_
   {
@@ -74,18 +70,14 @@
   static
   Translator  key_translators[] =
   {
-    { VK_BACK,      grKeyBackSpace },
-    { VK_TAB,       grKeyTab       },
-    { VK_RETURN,    grKeyReturn    },
-    { VK_ESCAPE,    grKeyEsc       },
+    { VK_PRIOR,     grKeyPageUp    },
+    { VK_NEXT,      grKeyPageDown  },
+    { VK_END,       grKeyEnd       },
     { VK_HOME,      grKeyHome      },
     { VK_LEFT,      grKeyLeft      },
     { VK_UP,        grKeyUp        },
     { VK_RIGHT,     grKeyRight     },
     { VK_DOWN,      grKeyDown      },
-    { VK_PRIOR,     grKeyPageUp    },
-    { VK_NEXT,      grKeyPageDown  },
-    { VK_END,       grKeyEnd       },
     { VK_F1,        grKeyF1        },
     { VK_F2,        grKeyF2        },
     { VK_F3,        grKeyF3        },
@@ -100,15 +92,16 @@
     { VK_F12,       grKeyF12       }
   };
 
-  static ATOM  ourAtom;
-
   typedef struct grWin32SurfaceRec_
   {
     grSurface     root;
+    DWORD         host;
     HWND          window;
-    LPBITMAPINFO  pbmi;
-    char          bmi[ sizeof(BITMAPINFO) + 256*sizeof(RGBQUAD) ];
-    grBitmap      bgrBitmap;  /* windows wants data in BGR format !! */
+    HICON         sIcon;
+    HICON         bIcon;
+    BITMAPINFOHEADER  bmiHeader;
+    RGBQUAD           bmiColors[256];
+    grBitmap      shadow_bitmap;  /* windows wants 24-bit BGR format !! */
 #ifdef SWIZZLE
     grBitmap      swizzle_bitmap;
 #endif
@@ -125,10 +118,16 @@ gr_win32_surface_done( grWin32Surface*  surface )
     DestroyWindow ( surface->window );
     PostMessage( surface->window, WM_QUIT, 0, 0 );
   }
+
+  DestroyIcon( surface->sIcon );
+  DestroyIcon( surface->bIcon );
+  if ( surface->root.bitmap.mode == gr_pixel_mode_rgb24 )
+  {
 #ifdef SWIZZLE
-  grDoneBitmap( &surface->swizzle_bitmap );
+    grDoneBitmap( &surface->swizzle_bitmap );
 #endif
-  grDoneBitmap( &surface->bgrBitmap );
+    grDoneBitmap( &surface->shadow_bitmap );
+  }
   grDoneBitmap( &surface->root.bitmap );
 }
 
@@ -143,11 +142,10 @@ gr_win32_surface_refresh_rectangle(
 {
   int        delta;
   RECT       rect;
-  HANDLE     window = surface->window;
   grBitmap*  bitmap = &surface->root.bitmap;
 
   LOG(( "gr_win32_surface_refresh_rectangle: ( %p, %d, %d, %d, %d )\n",
-        (long)surface, x, y, w, h ));
+        surface->root.bitmap.buffer, x, y, w, h ));
 
   /* clip update rectangle */
 
@@ -180,6 +178,7 @@ gr_win32_surface_refresh_rectangle(
   rect.bottom = y + h;
 
 #ifdef SWIZZLE
+  if ( bitmap->mode == gr_pixel_mode_rgb24 )
   {
     grBitmap*  swizzle = &surface->swizzle_bitmap;
 
@@ -193,12 +192,13 @@ gr_win32_surface_refresh_rectangle(
   }
 #endif
 
-  /* copy to BGR buffer */
+  /* copy the buffer */
+  if ( bitmap->mode == gr_pixel_mode_rgb24 )
   {
     unsigned char*  read_line   = (unsigned char*)bitmap->buffer;
     int             read_pitch  = bitmap->pitch;
-    unsigned char*  write_line  = (unsigned char*)surface->bgrBitmap.buffer;
-    int             write_pitch = surface->bgrBitmap.pitch;
+    unsigned char*  write_line  = (unsigned char*)surface->shadow_bitmap.buffer;
+    int             write_pitch = surface->shadow_bitmap.pitch;
 
     if ( read_pitch < 0 )
       read_line -= ( bitmap->rows - 1 ) * read_pitch;
@@ -206,45 +206,30 @@ gr_win32_surface_refresh_rectangle(
     if ( write_pitch < 0 )
       write_line -= ( bitmap->rows - 1 ) * write_pitch;
 
-    read_line  += y * read_pitch;
-    write_line += y * write_pitch;
+    read_line  += y * read_pitch  + 3 * x;
+    write_line += y * write_pitch + 3 * x;
 
-    if ( bitmap->mode == gr_pixel_mode_gray )
+    for ( ; h > 0; h-- )
     {
-      read_line  += x;
-      write_line += x;
-      for ( ; h > 0; h-- )
+      unsigned char*  read       = read_line;
+      unsigned char*  read_limit = read + 3 * w;
+      unsigned char*  write      = write_line;
+
+      /* convert RGB to BGR */
+      for ( ; read < read_limit; read += 3, write += 3 )
       {
-        memcpy( write_line, read_line, w );
-
-        read_line  += read_pitch;
-        write_line += write_pitch;
+        write[0] = read[2];
+        write[1] = read[1];
+        write[2] = read[0];
       }
-    }
-    else
-    {
-      read_line  += 3 * x;
-      write_line += 3 * x;
-      for ( ; h > 0; h-- )
-      {
-        unsigned char*  read       = read_line;
-        unsigned char*  read_limit = read + 3 * w;
-        unsigned char*  write      = write_line;
 
-        for ( ; read < read_limit; read += 3, write += 3 )
-        {
-          write[0] = read[2];
-          write[1] = read[1];
-          write[2] = read[0];
-        }
-
-        read_line  += read_pitch;
-        write_line += write_pitch;
-      }
+      read_line  += read_pitch;
+      write_line += write_pitch;
     }
   }
 
-  InvalidateRect( window, &rect, FALSE );
+  InvalidateRect( surface->window, &rect, FALSE );
+  UpdateWindow( surface->window );
 }
 
 
@@ -256,19 +241,57 @@ gr_win32_surface_set_title( grWin32Surface*  surface,
 }
 
 
+static int
+gr_win32_surface_set_icon( grWin32Surface*  surface,
+                           grBitmap*        icon )
+{
+  int     s[] = { GetSystemMetrics( SM_CYSMICON ),
+                  GetSystemMetrics( SM_CYICON ) };
+  WPARAM  wParam;
+  HICON   hIcon;
+
+  if ( !icon )
+    return s[1];
+  else if ( icon->mode != gr_pixel_mode_rgb32 )
+    return 0;
+  else if ( icon->rows == s[0] )
+    wParam = ICON_SMALL;
+  else if ( icon->rows == s[1] )
+    wParam = ICON_BIG;
+  else
+    return 0;
+
+  hIcon = CreateIcon( NULL, icon->width, icon->rows, 1, 32,
+                      NULL, icon->buffer );
+
+  PostMessage( surface->window, WM_SETICON, wParam, (LPARAM)hIcon );
+
+  switch( wParam )
+  {
+  case ICON_SMALL:
+    surface->sIcon = hIcon;
+    return 0;
+  case ICON_BIG:
+    surface->bIcon = hIcon;
+    return s[0];
+  default:
+    return 0;  /* should not happen */
+  }
+}
+
+
 /*
  * set graphics mode
  * and create the window class and the message handling.
  */
 
 
-static grWin32Surface*
+static int
 gr_win32_surface_resize( grWin32Surface*  surface,
                          int              width,
                          int              height )
 {
   grBitmap*       bitmap = &surface->root.bitmap;
-  LPBITMAPINFO    pbmi = surface->pbmi;
 
   /* resize root bitmap */
   if ( grNewBitmap( bitmap->mode,
@@ -277,35 +300,34 @@ gr_win32_surface_resize( grWin32Surface*  surface,
                     height,
                     bitmap ) )
     return 0;
-  bitmap->pitch = -bitmap->pitch;
 
   /* resize BGR shadow bitmap */
-  if ( grNewBitmap( bitmap->mode,
-                    bitmap->grays,
-                    width,
-                    height,
-                    &surface->bgrBitmap ) )
-    return 0;
-  surface->bgrBitmap.pitch = -surface->bgrBitmap.pitch;
-
-#ifdef SWIZZLE
   if ( bitmap->mode == gr_pixel_mode_rgb24 )
   {
+    if ( grNewBitmap( bitmap->mode,
+                      bitmap->grays,
+                      width,
+                      height,
+                      &surface->shadow_bitmap ) )
+      return 0;
+
+#ifdef SWIZZLE
     if ( grNewBitmap( bitmap->mode,
                       bitmap->grays,
                       width,
                       height,
                       &surface->swizzle_bitmap ) )
       return 0;
-    surface->swizzle_bitmap.pitch = -surface->swizzle_bitmap.pitch;
-  }
 #endif
+  }
+  else
+    surface->shadow_bitmap.buffer = bitmap->buffer;
 
   /* update the header to appropriate values */
-  pbmi->bmiHeader.biWidth  = width;
-  pbmi->bmiHeader.biHeight = height;
+  surface->bmiHeader.biWidth  = width;
+  surface->bmiHeader.biHeight = -height;
 
-  return surface;
+  return 1;
 }
 
 static void
@@ -314,15 +336,17 @@ gr_win32_surface_listen_event( grWin32Surface*  surface,
                                grEvent*         grevent )
 {
   MSG     msg;
-  HANDLE  window = surface->window;
 
   event_mask=event_mask;  /* unused parameter */
 
-  while (GetMessage( &msg, NULL, 0, 0 ) > 0)
+  while ( GetMessage( &msg, (HWND)-1, 0, 0 ) > 0 )
   {
     switch ( msg.message )
     {
     case WM_RESIZE:
+      while( PeekMessage( &msg, (HWND)-1, WM_RESIZE, WM_RESIZE, PM_REMOVE ) )
+        continue;
+
       {
         int  width  = LOWORD(msg.lParam);
         int  height = HIWORD(msg.lParam);
@@ -340,177 +364,203 @@ gr_win32_surface_listen_event( grWin32Surface*  surface,
       }
       break;
 
-    case WM_KEYDOWN:
-    case WM_SYSKEYDOWN:
-      {
-        Translator*  trans = key_translators;
-        Translator*  limit = trans + sizeof( key_translators ) /
-                                     sizeof( key_translators[0] );
-        for ( ; trans < limit; trans++ )
-          if ( msg.wParam == trans->winkey )
-          {
-            grevent->type = gr_event_key;
-            grevent->key  = trans->grkey;
-            LOG(( "KeyPress: VK = 0x%02x\n", msg.wParam ));
-            return;
-          }
-      }
-      break;
-
     case WM_CHAR:
+    case WM_GR_KEY:
       {
         grevent->type = gr_event_key;
         grevent->key  = msg.wParam;
-        LOG(( isprint( msg.wParam ) ? "KeyPress: Char = '%c'\n"
+        LOG(( (msg.wParam <= 256 && isprint( msg.wParam ))
+                                    ? "KeyPress: Char = '%c'\n"
                                     : "KeyPress: Char = <%02x>\n",
               msg.wParam ));
         return;
       }
       break;
     }
-
-    TranslateMessage( &msg );
-    DispatchMessage( &msg );
   }
 }
 
-static grWin32Surface*
+
+DWORD WINAPI Window_ThreadProc( LPVOID lpParameter )
+{
+  grWin32Surface*  surface = (grWin32Surface*)lpParameter;
+  DWORD            style   = WS_OVERLAPPEDWINDOW;
+  RECT             WndRect;
+  MSG              msg;
+
+  WndRect.left   = 0;
+  WndRect.top    = 0;
+  WndRect.right  = surface->root.bitmap.width;
+  WndRect.bottom = surface->root.bitmap.rows;
+
+  AdjustWindowRect( &WndRect, style, FALSE );
+
+  surface->window = CreateWindow(
+       /* LPCSTR lpszClassName;    */ "FreeTypeTestGraphicDriver",
+       /* LPCSTR lpszWindowName;   */ "FreeType Test Graphic Driver",
+       /* DWORD dwStyle;           */  style,
+       /* int x;                   */  CW_USEDEFAULT,
+       /* int y;                   */  CW_USEDEFAULT,
+       /* int nWidth;              */  WndRect.right - WndRect.left,
+       /* int nHeight;             */  WndRect.bottom - WndRect.top,
+       /* HWND hwndParent;         */  HWND_DESKTOP,
+       /* HMENU hmenu;             */  NULL,
+       /* HINSTANCE hinst;         */  GetModuleHandle( NULL ),
+       /* void FAR* lpvParam;      */  surface );
+
+  PostThreadMessage( surface->host, WM_STATUS, (WPARAM)surface->window, 0 );
+
+  if ( surface->window == NULL )
+    return -1;
+
+  ShowWindow( surface->window, SW_SHOWNORMAL );
+
+  while ( GetMessage( &msg, NULL, 0, 0 ) > 0 )
+  {
+    TranslateMessage( &msg );
+    DispatchMessage( &msg );
+  }
+
+  LOG(("Window thread done.\n"));
+  return 0;
+}
+
+
+static int
 gr_win32_surface_init( grWin32Surface*  surface,
                        grBitmap*        bitmap )
 {
-  static RGBQUAD  black = {    0,    0,    0, 0 };
-  static RGBQUAD  white = { 0xFF, 0xFF, 0xFF, 0 };
-  LPBITMAPINFO    pbmi;
+  MSG  msg;
 
-  LOG(( "Win32: init_surface( %p, %p )\n", surface, bitmap ));
 
-  LOG(( "       -- input bitmap =\n" ));
-  LOG(( "       --   mode   = %d\n", bitmap->mode ));
-  LOG(( "       --   grays  = %d\n", bitmap->grays ));
-  LOG(( "       --   width  = %d\n", bitmap->width ));
-  LOG(( "       --   height = %d\n", bitmap->rows ));
-
-  /* create the bitmap - under Win32, we support all modes as the GDI */
-  /* handles all conversions automatically..                          */
-  if ( grNewBitmap( bitmap->mode,
-                    bitmap->grays,
-                    bitmap->width,
-                    bitmap->rows,
-                    bitmap ) )
-    return 0;
-  bitmap->pitch = -bitmap->pitch;
-
-  /* allocate the BGR shadow bitmap */
-  if ( grNewBitmap( bitmap->mode,
-                    bitmap->grays,
-                    bitmap->width,
-                    bitmap->rows,
-                    &surface->bgrBitmap ) )
-    return 0;
-  surface->bgrBitmap.pitch = -surface->bgrBitmap.pitch;
-
-#ifdef SWIZZLE
-  if ( bitmap->mode == gr_pixel_mode_rgb24 )
+  /* Set default mode */
+  if ( bitmap->mode == gr_pixel_mode_none )
   {
-    if ( grNewBitmap( bitmap->mode,
-                      bitmap->grays,
-                      bitmap->width,
-                      bitmap->rows,
-                      &surface->swizzle_bitmap ) )
-      return 0;
-    surface->swizzle_bitmap.pitch = -surface->swizzle_bitmap.pitch;
+    HDC  hDC;
+    int  bpp;
+
+    hDC = GetDC( NULL );
+    bpp = GetDeviceCaps( hDC, BITSPIXEL ) * GetDeviceCaps( hDC, PLANES );
+    ReleaseDC( NULL, hDC );
+
+    switch ( bpp )
+    {
+    case 8:
+      surface->root.bitmap.mode  = gr_pixel_mode_gray;
+      surface->root.bitmap.grays = 256;
+      break;
+    case 16:
+      surface->root.bitmap.mode = gr_pixel_mode_rgb565;
+      break;
+    case 24:
+      surface->root.bitmap.mode = gr_pixel_mode_rgb24;
+      break;
+    case 32:
+    default:
+      surface->root.bitmap.mode = gr_pixel_mode_rgb32;
+    }
+  }
+  else
+  {
+    surface->root.bitmap.mode  = bitmap->mode;
+    surface->root.bitmap.grays = bitmap->grays;
   }
-#endif
 
-  LOG(( "       -- output bitmap =\n" ));
-  LOG(( "       --   mode   = %d\n", bitmap->mode ));
-  LOG(( "       --   grays  = %d\n", bitmap->grays ));
-  LOG(( "       --   width  = %d\n", bitmap->width ));
-  LOG(( "       --   height = %d\n", bitmap->rows ));
+  surface->bmiHeader.biSize   = sizeof( BITMAPINFOHEADER );
+  surface->bmiHeader.biPlanes = 1;
 
-  /* find some memory for the bitmap header */
-  surface->pbmi = pbmi = (LPBITMAPINFO) surface->bmi;
-
-  /* initialize the header to appropriate values */
-  memset( pbmi, 0, sizeof ( BITMAPINFO ) + sizeof ( RGBQUAD ) * 256 );
-
-  pbmi->bmiHeader.biSize   = sizeof ( BITMAPINFOHEADER );
-  pbmi->bmiHeader.biWidth  = bitmap->width;
-  pbmi->bmiHeader.biHeight = bitmap->rows;
-  pbmi->bmiHeader.biPlanes = 1;
-
-  switch ( bitmap->mode )
+  switch ( surface->root.bitmap.mode )
   {
   case gr_pixel_mode_mono:
-    pbmi->bmiHeader.biBitCount = 1;
-    pbmi->bmiColors[0] = white;
-    pbmi->bmiColors[1] = black;
-    break;
+    surface->bmiHeader.biBitCount = 1;
+    {
+      RGBQUAD  white = { 0xFF, 0xFF, 0xFF, 0 };
+      RGBQUAD  black = {    0,    0,    0, 0 };
 
-  case gr_pixel_mode_rgb24:
-    pbmi->bmiHeader.biBitCount    = 24;
-    pbmi->bmiHeader.biCompression = BI_RGB;
+      surface->bmiColors[0] = white;
+      surface->bmiColors[1] = black;
+    }
     break;
 
   case gr_pixel_mode_gray:
-    pbmi->bmiHeader.biBitCount = 8;
-    pbmi->bmiHeader.biClrUsed  = bitmap->grays;
+    surface->bmiHeader.biBitCount = 8;
+    surface->bmiHeader.biClrUsed  = surface->root.bitmap.grays;
     {
-      int   count = bitmap->grays;
-      int   x;
-      RGBQUAD*  color = pbmi->bmiColors;
+      int       x, count = surface->root.bitmap.grays - 1;
+      RGBQUAD*  color = surface->bmiColors;
 
-      for ( x = 0; x < count; x++, color++ )
+      for ( x = 0; x <= count; x++, color++ )
       {
         color->rgbRed   =
         color->rgbGreen =
-        color->rgbBlue  = (unsigned char)(x*255/(count-1));
+        color->rgbBlue  = (BYTE)( x * 255 / count );
         color->rgbReserved = 0;
       }
     }
     break;
 
+  case gr_pixel_mode_rgb32:
+    surface->bmiHeader.biBitCount    = 32;
+    surface->bmiHeader.biCompression = BI_RGB;
+    break;
+
+  case gr_pixel_mode_rgb24:
+    surface->bmiHeader.biBitCount    = 24;
+    surface->bmiHeader.biCompression = BI_RGB;
+    break;
+
+  case gr_pixel_mode_rgb555:
+    surface->bmiHeader.biBitCount    = 16;
+    surface->bmiHeader.biCompression = BI_RGB;
+    break;
+
+  case gr_pixel_mode_rgb565:
+    surface->bmiHeader.biBitCount    = 16;
+    surface->bmiHeader.biCompression = BI_BITFIELDS;
+    {
+       LPDWORD  mask = (LPDWORD)surface->bmiColors;
+
+       mask[0] = 0xF800;
+       mask[1] = 0x07E0;
+       mask[2] = 0x001F;
+    }
+    break;
+
   default:
     return 0;         /* Unknown mode */
   }
 
-  {
-    DWORD  style = WS_OVERLAPPEDWINDOW;
-    RECT   WndRect;
+  /* allocate buffers */
+  if ( !gr_win32_surface_resize( surface, bitmap->width, bitmap->rows ) )
+    goto Fail;
 
-    WndRect.left   = 0;
-    WndRect.top    = 0;
-    WndRect.right  = bitmap->width;
-    WndRect.bottom = bitmap->rows;
+  /* set up the main message queue and spin off the window thread */
+  PeekMessage( &msg, (HWND)-1, WM_USER, WM_USER, PM_NOREMOVE );
+  surface->host = GetCurrentThreadId();
 
-    AdjustWindowRect( &WndRect, style, FALSE );
+  CreateThread( NULL, 0, Window_ThreadProc, (LPVOID)surface, 0, NULL );
 
-    surface->window = CreateWindow(
-            /* LPCSTR lpszClassName;    */ "FreeTypeTestGraphicDriver",
-            /* LPCSTR lpszWindowName;   */ "FreeType Test Graphic Driver",
-            /* DWORD dwStyle;           */  style,
-            /* int x;                   */  CW_USEDEFAULT,
-            /* int y;                   */  CW_USEDEFAULT,
-            /* int nWidth;              */  WndRect.right - WndRect.left,
-            /* int nHeight;             */  WndRect.bottom - WndRect.top,
-            /* HWND hwndParent;         */  HWND_DESKTOP,
-            /* HMENU hmenu;             */  0,
-            /* HINSTANCE hinst;         */  GetModuleHandle( NULL ),
-            /* void FAR* lpvParam;      */  surface );
-  }
+  /* listen if window is created */
+  if ( GetMessage ( &msg, (HWND)-1, WM_STATUS, WM_STATUS ) < 0 ||
+       !msg.wParam )
+    goto Fail;
 
-  if ( surface->window == 0 )
-    return  0;
-
-  ShowWindow( surface->window, SW_SHOWNORMAL );
-
-  surface->root.bitmap       = *bitmap;
+  /* wrap up */
   surface->root.done         = (grDoneSurfaceFunc) gr_win32_surface_done;
   surface->root.refresh_rect = (grRefreshRectFunc) gr_win32_surface_refresh_rectangle;
   surface->root.set_title    = (grSetTitleFunc)    gr_win32_surface_set_title;
+  surface->root.set_icon     = (grSetIconFunc)     gr_win32_surface_set_icon;
   surface->root.listen_event = (grListenEventFunc) gr_win32_surface_listen_event;
 
-  return surface;
+  LOG(( "Surface initialized: %dx%dx%d\n",
+        surface->root.bitmap.width, surface->root.bitmap.rows,
+        surface->bmiHeader.biBitCount ));
+
+  return 1;
+
+Fail:
+  gr_win32_surface_done( surface );
+  return 0;
 }
 
 
@@ -521,7 +571,7 @@ gr_win32_surface_init( grWin32Surface*  surface,
 LRESULT CALLBACK Message_Process( HWND handle, UINT mess,
                                   WPARAM wParam, LPARAM lParam )
   {
-    grWin32Surface*  surface = NULL;
+    grWin32Surface*  surface;
 
     if ( mess == WM_CREATE )
     {
@@ -531,62 +581,79 @@ LRESULT CALLBACK Message_Process( HWND handle, UINT mess,
       /*                                                               */
       surface = ((LPCREATESTRUCT)lParam)->lpCreateParams;
 
-      SetProp( handle, (LPCSTR)(LONG)ourAtom, surface );
+      SetWindowLongPtr( handle, GWLP_USERDATA, (LONG_PTR)surface );
     }
     else
     {
       /* for other calls, we retrieve the surface handle from the window */
       /* property.. ugly, isn't it ??                                    */
       /*                                                                 */
-      surface = (grWin32Surface*) GetProp( handle, (LPCSTR)(LONG)ourAtom );
+      surface = (grWin32Surface*)GetWindowLongPtr( handle, GWLP_USERDATA );
     }
 
     switch( mess )
     {
     case WM_CLOSE:
       /* warn the main thread to quit if it didn't know */
-      PostMessage( handle, WM_CHAR, (WPARAM)grKeyEsc, 0 );
-      return 0;
+      PostThreadMessage( surface->host, WM_GR_KEY, (WPARAM)grKeyEsc, 0 );
+      break;
 
     case WM_SIZE:
       if ( wParam == SIZE_RESTORED || wParam == SIZE_MAXIMIZED )
-        PostMessage( handle, WM_RESIZE, wParam, lParam );
+        PostThreadMessage( surface->host, WM_RESIZE, wParam, lParam );
       break;
 
-    case WM_EXITSIZEMOVE:
+    case WM_KEYDOWN:
+    case WM_SYSKEYDOWN:
       {
-        RECT  WndRect;
+        Translator*  trans = key_translators;
+        Translator*  limit = trans + sizeof( key_translators ) /
+                                     sizeof( key_translators[0] );
 
-        GetClientRect( handle, &WndRect );
-        PostMessage( handle, WM_RESIZE, SIZE_RESTORED,
-                     MAKELPARAM( WndRect.right, WndRect.bottom ) );
+        for ( ; trans < limit; trans++ )
+          if ( wParam == trans->winkey )
+          {
+            /* repost to the main thread */
+            PostThreadMessage( surface->host, WM_GR_KEY, trans->grkey, 0 );
+            LOG(( "KeyPress: VK = 0x%02x\n", wParam ));
+            break;
+          }
       }
       break;
 
+    case WM_CHAR:
+      /* repost to the main thread */
+      PostThreadMessage( surface->host, mess, wParam, lParam );
+      break;
+
     case WM_PAINT:
       {
-      HDC           hDC;
-      PAINTSTRUCT   ps;
-      LPBITMAPINFO  pbmi = surface->pbmi;
+        HDC          hDC;
+        PAINTSTRUCT  ps;
 
-      hDC   = BeginPaint ( handle, &ps );
-      if ( pbmi )
-      {
-        SetDIBitsToDevice( hDC, 0, 0,
-                           pbmi->bmiHeader.biWidth,
-                           pbmi->bmiHeader.biHeight,
-                           0, 0, 0,
-                           pbmi->bmiHeader.biHeight,
-                           surface->bgrBitmap.buffer,
-                           pbmi,
+        hDC = BeginPaint ( handle, &ps );
+
+        LOG(( "WM_PAINT : ( %d %d %d %d )\n",
+              ps.rcPaint.left,  ps.rcPaint.top,
+              ps.rcPaint.right, ps.rcPaint.bottom ));
+
+        SetDIBitsToDevice( hDC,
+                           ps.rcPaint.left,
+                           ps.rcPaint.top,
+                           ps.rcPaint.right - ps.rcPaint.left,
+                           ps.rcPaint.bottom - ps.rcPaint.top,
+                           ps.rcPaint.left,
+                           surface->root.bitmap.rows - ps.rcPaint.bottom,
+                           0, surface->root.bitmap.rows,
+                           surface->shadow_bitmap.buffer,
+                           (LPBITMAPINFO)&surface->bmiHeader,
                            DIB_RGB_COLORS );
+        EndPaint ( handle, &ps );
       }
-      EndPaint ( handle, &ps );
-      return 0;
-      }
+      break;
 
     default:
-       return DefWindowProc( handle, mess, wParam, lParam );
+      return DefWindowProc( handle, mess, wParam, lParam );
     }
     return 0;
   }
@@ -599,34 +666,25 @@ LRESULT CALLBACK Message_Process( HWND handle, UINT mess,
       /* WNDPROC lpfnWndProc  */ Message_Process,
       /* int     cbClsExtra   */ 0,
       /* int     cbWndExtra   */ 0,
-      /* HANDLE  hInstance    */ 0,
-      /* HICON   hIcon        */ 0,
-      /* HCURSOR hCursor      */ 0,
-      /* HBRUSH  hbrBackground*/ 0,
+      /* HANDLE  hInstance    */ GetModuleHandle( NULL ),
+      /* HICON   hIcon        */ NULL,
+      /* HCURSOR hCursor      */ LoadCursor( NULL, IDC_ARROW),
+      /* HBRUSH  hbrBackground*/ NULL,
       /* LPCTSTR lpszMenuName */ NULL,
       /* LPCTSTR lpszClassName*/ "FreeTypeTestGraphicDriver"
     };
 
     /* register window class */
-
-    ourClass.hInstance    = GetModuleHandle( NULL );
-    ourClass.hIcon        = LoadIcon(0, IDI_APPLICATION);
-    ourClass.hCursor      = LoadCursor(0, IDC_ARROW);
-    ourClass.hbrBackground= GetStockObject(BLACK_BRUSH);
-
-    if ( RegisterClass(&ourClass) == 0 )
+    if ( RegisterClass( &ourClass ) == 0 )
       return -1;
 
-    /* add global atom */
-    ourAtom = GlobalAddAtom( "FreeType.Surface" );
-
     return 0;
   }
 
   static void
   gr_win32_device_done( void )
   {
-    GlobalDeleteAtom( ourAtom );
+    /* Nothing to do. */
   }
 
 
