@@ -2,7 +2,7 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality TrueType renderer.  */
 /*                                                                          */
-/*  Copyright (C) 1996-2020 by                                              */
+/*  Copyright (C) 1996-2023 by                                              */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
 /*                                                                          */
@@ -13,14 +13,16 @@
 /****************************************************************************/
 
 #include <ft2build.h>
-#include FT_FREETYPE_H
-#include FT_FONT_FORMATS_H
-#include FT_MODULE_H
-#include FT_DRIVER_H
-#include FT_MULTIPLE_MASTERS_H
+#include <freetype/freetype.h>
+
+#include <freetype/ftdriver.h>
+#include <freetype/ftfntfmt.h>
+#include <freetype/ftmm.h>
+#include <freetype/ftmodapi.h>
 
 #include "common.h"
 #include "mlgetopt.h"
+#include "strbuf.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -33,8 +35,10 @@
 #define  DIM_X   640
 #define  DIM_Y   480
 
+#define  HEADER_HEIGHT  12
+
 #define  MAXPTSIZE    500               /* dtp */
-#define  MAX_MM_AXES    6
+#define  MAX_MM_AXES   16
 
   /* definitions in ftcommon.c */
   unsigned int
@@ -106,8 +110,9 @@
   static FT_MM_Var    *multimaster   = NULL;
   static FT_Fixed      design_pos   [MAX_MM_AXES];
   static FT_Fixed      requested_pos[MAX_MM_AXES];
-  static unsigned int  requested_cnt = 0;
-  static unsigned int  used_num_axis = 0;
+  static unsigned int  requested_cnt =  0;
+  static unsigned int  used_num_axis =  0;
+  static int           increment     = 20;  /* for axes */
 
   /*
    * We use the following arrays to support both the display of all axes and
@@ -257,7 +262,7 @@
         }
       }
 
-      num_shown_axes = idx + 1;
+      num_shown_axes = (unsigned int)( idx + 1 );
     }
     else
     {
@@ -266,7 +271,7 @@
 
       /* show all axes */
       for ( i = 0; i < used_num_axis; i++ )
-        shown_axes[i] = i;
+        shown_axes[i] = (int)i;
 
       num_shown_axes = used_num_axis;
     }
@@ -277,12 +282,10 @@
   static void
   Clear_Display( void )
   {
-    long  bitmap_size = (long)bit->pitch * bit->rows;
-
-
-    if ( bitmap_size < 0 )
-      bitmap_size = -bitmap_size;
-    memset( bit->buffer, 0, (unsigned long)bitmap_size );
+    /* fast black background */
+    memset( bit->buffer, 0, (size_t)bit->rows *
+                            (size_t)( bit->pitch < 0 ? -bit->pitch
+                                                     : bit->pitch ) );
   }
 
 
@@ -345,8 +348,8 @@
     x_top = x_offset + glyph->bitmap_left;
     y_top = y_offset - glyph->bitmap_top;
 
-    grBlitGlyphToBitmap( bit, &bit3,
-                         x_top, y_top, fore_color );
+    grBlitGlyphToSurface( surface, &bit3,
+                          x_top, y_top, fore_color );
 
     return 0;
   }
@@ -384,12 +387,12 @@
   Render_All( unsigned int  first_glyph,
               int           pt_size )
   {
-    FT_F26Dot6    start_x, start_y, step_y, x, y;
+    int           start_x, start_y, step_y, x, y, w;
     unsigned int  i;
 
 
-    start_x = 4;
-    start_y = pt_size + ( num_shown_axes > MAX_MM_AXES / 2 ? 52 : 44 );
+    start_x = 20 * 8;
+    start_y = pt_size + HEADER_HEIGHT * 3;
 
     step_y = size->metrics.y_ppem + 10;
 
@@ -419,11 +422,8 @@
         }
 #endif
 
-        Render_Glyph( x, y );
-
-        x += ( ( glyph->metrics.horiAdvance + 32 ) >> 6 ) + 1;
-
-        if ( x + size->metrics.x_ppem > bit->width )
+        w = ( ( glyph->metrics.horiAdvance + 32 ) >> 6 ) + 1;
+        if ( x + w > bit->width - 4 )
         {
           x  = start_x;
           y += step_y;
@@ -431,6 +431,9 @@
           if ( y >= bit->rows )
             return FT_Err_Ok;
         }
+
+        Render_Glyph( x, y );
+        x += w;
       }
       else
         Fail++;
@@ -452,8 +455,8 @@
     const unsigned char*  p;
 
 
-    start_x = 4;
-    start_y = pt_size + ( num_shown_axes > MAX_MM_AXES / 2 ? 52 : 44 );
+    start_x = 20 * 8;
+    start_y = pt_size + HEADER_HEIGHT * 3;
 
     step_y = size->metrics.y_ppem + 10;
 
@@ -525,9 +528,11 @@
     FT_Library_Version( library, &major, &minor, &patch );
 
     if ( patch )
-      sprintf( version, "%d.%d.%d", major, minor, patch );
+      snprintf( version, sizeof ( version ),
+                "%d.%d.%d", major, minor, patch );
     else
-      sprintf( version, "%d.%d", major, minor );
+      snprintf( version, sizeof ( version ),
+                "%d.%d", major, minor );
 
     Clear_Display();
     grSetLineHeight( 10 );
@@ -535,9 +540,10 @@
     grSetMargin( 2, 1 );
     grGotobitmap( bit );
 
-    sprintf( buf,
-             "FreeType MM Glyph Viewer - part of the FreeType %s test suite",
-             version );
+    snprintf( buf, sizeof ( buf ),
+              "FreeType MM Glyph Viewer -"
+                " part of the FreeType %s test suite",
+              version );
 
     grWriteln( buf );
     grLn();
@@ -546,31 +552,32 @@
     grLn();
     grWriteln( "Use the following keys:");
     grLn();
-    grWriteln( "?           display this help screen" );
-    grWriteln( "A           toggle axis grouping" );
-    grWriteln( "a           toggle anti-aliasing" );
-    grWriteln( "h           toggle outline hinting" );
-    grWriteln( "b           toggle embedded bitmaps" );
+    grWriteln( "F1, ?       display this help screen" );
+    grWriteln( "q, ESC      quit ftmulti" );
+    grWriteln( "F2          toggle axis grouping" );
+    grWriteln( "F3          toggle outline hinting" );
+    grWriteln( "F4          toggle embedded bitmaps" );
+    grWriteln( "F5          toggle anti-aliasing" );
+    grWriteln( "F6          cycle through hinting engines (if available)" );
+    grLn();
     grWriteln( "space       toggle rendering mode" );
     grLn();
-    grWriteln( "p, n        previous/next font" );
-    grLn();
-    grWriteln( "H           cycle through hinting engines (if available)" );
+    grWriteln( ", .         previous/next font" );
     grLn();
     grWriteln( "Up, Down    change pointsize by 1 unit" );
     grWriteln( "PgUp, PgDn  change pointsize by 10 units" );
     grLn();
     grWriteln( "Left, Right adjust index by 1" );
-    grWriteln( "F7, F8      adjust index by 10" );
-    grWriteln( "F9, F10     adjust index by 100" );
-    grWriteln( "F11, F12    adjust index by 1000" );
+    grWriteln( "F7, F8      adjust index by 16" );
+    grWriteln( "F9, F10     adjust index by 256" );
+    grWriteln( "F11, F12    adjust index by 4096" );
     grLn();
-    grWriteln( "F1, F2      adjust first axis by 1/50th of its range" );
-    grWriteln( "F3, F4      adjust second axis by 1/50th of its range" );
-    grWriteln( "F5, F6      adjust third axis by 1/50th of its range" );
-    grWriteln( "1, 2        adjust fourth axis by 1/50th of its range" );
-    grWriteln( "3, 4        adjust fifth axis by 1/50th of its range" );
-    grWriteln( "5, 6        adjust sixth axis by 1/50th of its range" );
+    grWriteln( "a, A        adjust axis 0" );
+    grWriteln( "b, B        adjust axis 1" );
+    grWriteln( "..." );
+    grWriteln( "p, P        adjust axis 16" );
+    grLn();
+    grWriteln( "-, +        adjust axis range increment" );
     grLn();
     grWriteln( "Axes marked with an asterisk are hidden." );
     grLn();
@@ -603,45 +610,50 @@
     unsigned int  axis;
 
 
+    grRefreshSurface( surface );
     grListenSurface( surface, 0, &event );
 
+    if ( event.type == gr_event_resize )
+      return 1;
+
     switch ( event.key )
     {
     case grKeyEsc:            /* ESC or q */
     case grKEY( 'q' ):
       return 0;
 
+    case grKeyF1:
     case grKEY( '?' ):
       Help();
       break;
 
     /* mode keys */
 
-    case grKEY( 'A' ):
+    case grKeyF2:
       grouping = !grouping;
       new_header = grouping ? "axis grouping is now on"
                             : "axis grouping is now off";
       set_up_axes();
       break;
 
-    case grKEY( 'a' ):
+    case grKeyF5:
       antialias  = !antialias;
       new_header = antialias ? "anti-aliasing is now on"
                              : "anti-aliasing is now off";
       break;
 
-    case grKEY( 'b' ):
+    case grKeyF4:
       use_sbits  = !use_sbits;
       new_header = use_sbits
                      ? "embedded bitmaps are now used if available"
                      : "embedded bitmaps are now ignored";
       break;
 
-    case grKEY( 'n' ):
-    case grKEY( 'p' ):
+    case grKEY( ',' ):
+    case grKEY( '.' ):
       return (int)event.key;
 
-    case grKEY( 'h' ):
+    case grKeyF3:
       hinted     = !hinted;
       new_header = hinted ? "glyph hinting is now active"
                           : "glyph hinting is now ignored";
@@ -653,7 +665,7 @@
                                  : "rendering test text string";
       break;
 
-    case grKEY( 'H' ):
+    case grKeyF6:
       if ( !strcmp( font_format, "CFF" ) )
         FTDemo_Event_Cff_Hinting_Engine_Change( library,
                                                 &cff_hinting_engine,
@@ -670,66 +682,57 @@
         tt_interpreter_version_change();
       break;
 
-    /* MM related keys */
+    /* MM-related keys */
 
-    case grKeyF1:
-      i = -20;
-      axis = 0;
+    case grKEY( '+' ):
+      /* value 100 is arbitrary */
+      if ( increment < 100 )
+        increment += 1;
+      break;
+
+    case grKEY( '-' ):
+      if ( increment > 1 )
+        increment -= 1;
+      break;
+
+    case grKEY( 'a' ):
+    case grKEY( 'b' ):
+    case grKEY( 'c' ):
+    case grKEY( 'd' ):
+    case grKEY( 'e' ):
+    case grKEY( 'f' ):
+    case grKEY( 'g' ):
+    case grKEY( 'h' ):
+    case grKEY( 'i' ):
+    case grKEY( 'j' ):
+    case grKEY( 'k' ):
+    case grKEY( 'l' ):
+    case grKEY( 'm' ):
+    case grKEY( 'n' ):
+    case grKEY( 'o' ):
+    case grKEY( 'p' ):
+      i = -increment;
+      axis = event.key - 'a';
       goto Do_Axis;
 
-    case grKeyF2:
-      i = 20;
-      axis = 0;
-      goto Do_Axis;
-
-    case grKeyF3:
-      i = -20;
-      axis = 1;
-      goto Do_Axis;
-
-    case grKeyF4:
-      i = 20;
-      axis = 1;
-      goto Do_Axis;
-
-    case grKeyF5:
-      i = -20;
-      axis = 2;
-      goto Do_Axis;
-
-    case grKeyF6:
-      i = 20;
-      axis = 2;
-      goto Do_Axis;
-
-    case grKEY( '1' ):
-      i = -20;
-      axis = 3;
-      goto Do_Axis;
-
-    case grKEY( '2' ):
-      i = 20;
-      axis = 3;
-      goto Do_Axis;
-
-    case grKEY( '3' ):
-      i = -20;
-      axis = 4;
-      goto Do_Axis;
-
-    case grKEY( '4' ):
-      i = 20;
-      axis = 4;
-      goto Do_Axis;
-
-    case grKEY( '5' ):
-      i = -20;
-      axis = 5;
-      goto Do_Axis;
-
-    case grKEY( '6' ):
-      i = 20;
-      axis = 5;
+    case grKEY( 'A' ):
+    case grKEY( 'B' ):
+    case grKEY( 'C' ):
+    case grKEY( 'D' ):
+    case grKEY( 'E' ):
+    case grKEY( 'F' ):
+    case grKEY( 'G' ):
+    case grKEY( 'H' ):
+    case grKEY( 'I' ):
+    case grKEY( 'J' ):
+    case grKEY( 'K' ):
+    case grKEY( 'L' ):
+    case grKEY( 'M' ):
+    case grKEY( 'N' ):
+    case grKEY( 'O' ):
+    case grKEY( 'P' ):
+      i = increment;
+      axis = event.key - 'A';
       goto Do_Axis;
 
     /* scaling related keys */
@@ -761,27 +764,27 @@
       goto Do_Glyph;
 
     case grKeyF7:
-      i = -10;
+      i = -16;
       goto Do_Glyph;
 
     case grKeyF8:
-      i = 10;
+      i = 16;
       goto Do_Glyph;
 
     case grKeyF9:
-      i = -100;
+      i = -256;
       goto Do_Glyph;
 
     case grKeyF10:
-      i = 100;
+      i = 256;
       goto Do_Glyph;
 
     case grKeyF11:
-      i = -1000;
+      i = -4096;
       goto Do_Glyph;
 
     case grKeyF12:
-      i = 1000;
+      i = 4096;
       goto Do_Glyph;
 
     default:
@@ -867,7 +870,7 @@
 
 
   static void
-  usage( char*  execname )
+  usage( const char*  execname )
   {
     fprintf( stderr,
       "\n"
@@ -875,7 +878,7 @@
       "--------------------------------------------------------\n"
       "\n" );
     fprintf( stderr,
-      "Usage: %s [options] pt font ...\n"
+      "Usage: %s [options] [pt] font ...\n"
       "\n",
              execname );
     fprintf( stderr,
@@ -914,7 +917,6 @@
     int    old_ptsize, orig_ptsize, file;
     int    first_glyph = 0;
     int    XisSetup = 0;
-    char*  execname;
     int    option;
     int    file_loaded;
 
@@ -924,10 +926,9 @@
     unsigned int  versions[3] = { TT_INTERPRETER_VERSION_35,
                                   TT_INTERPRETER_VERSION_38,
                                   TT_INTERPRETER_VERSION_40 };
+    const char*   execname = ft_basename( argv[0] );
 
 
-    execname = ft_basename( argv[0] );
-
     /* Initialize engine */
     error = FT_Init_FreeType( &library );
     if ( error )
@@ -981,7 +982,7 @@
         break;
 
       case 'f':
-        first_glyph = atoi( optarg );
+        sscanf( optarg, "%i", &first_glyph );
         break;
 
       case 'h':
@@ -1026,13 +1027,18 @@
     argc -= optind;
     argv += optind;
 
-    if ( argc <= 1 )
+    if ( argc == 0 )
       usage( execname );
 
-    if ( sscanf( argv[0], "%d", &orig_ptsize ) != 1 )
+    if ( argc > 1 && sscanf( argv[0], "%d", &orig_ptsize ) == 1 )
+    {
+      argc--;
+      argv++;
+    }
+    else
       orig_ptsize = 64;
 
-    file = 1;
+    file = 0;
 
   NewFile:
     ptsize      = orig_ptsize;
@@ -1069,7 +1075,7 @@
     /* set the current position to the default of each axis */
     if ( multimaster->num_axis > MAX_MM_AXES )
     {
-      fprintf( stderr, "only handling first %d variation axes (of %d)\n",
+      fprintf( stderr, "only handling first %u variation axes (of %u)\n",
                        MAX_MM_AXES, multimaster->num_axis );
       used_num_axis = MAX_MM_AXES;
     }
@@ -1141,11 +1147,15 @@
 
     for ( ;; )
     {
-      int  key;
+      int     key;
+      StrBuf  header[1];
 
 
       Clear_Display();
 
+      strbuf_init( header, Header, sizeof ( Header ) );
+      strbuf_reset( header );
+
       if ( file_loaded >= 1 )
       {
         switch ( render_mode )
@@ -1158,10 +1168,10 @@
           Render_All( (unsigned int)Num, ptsize );
         }
 
-        sprintf( Header, "%.50s %.50s (file %.100s)",
-                         face->family_name,
-                         face->style_name,
-                         ft_basename( argv[file] ) );
+        strbuf_format( header, "%.50s %.50s (file %.100s)",
+                       face->family_name,
+                       face->style_name,
+                       ft_basename( argv[file] ) );
 
         if ( !new_header )
           new_header = Header;
@@ -1169,59 +1179,27 @@
         grWriteCellString( bit, 0, 0, new_header, fore_color );
         new_header = NULL;
 
-        sprintf( Header, "PS name: %s",
-                         FT_Get_Postscript_Name( face ) );
-        grWriteCellString( bit, 0, 16, Header, fore_color );
+        strbuf_reset( header );
+        strbuf_format( header, "PS name: %s",
+                       FT_Get_Postscript_Name( face ) );
+        grWriteCellString( bit, 0, 2 * HEADER_HEIGHT, Header, fore_color );
 
-        sprintf( Header, "axes:" );
+        strbuf_reset( header );
+        strbuf_add( header, "axes:" );
+        grWriteCellString( bit, 0, 3 * HEADER_HEIGHT, Header, fore_color );
+        for ( n = 0; n < num_shown_axes; n++ )
         {
-          unsigned int  limit = num_shown_axes > MAX_MM_AXES / 2
-                                  ? MAX_MM_AXES / 2
-                                  : num_shown_axes;
+          int  axis = shown_axes[n];
 
 
-          for ( n = 0; n < limit; n++ )
-          {
-            char  temp[100];
-            int   axis;
-
-
-            axis = shown_axes[n];
-
-            sprintf( temp, "  %.50s%s: %.02f",
-                           multimaster->axis[axis].name,
-                           hidden[axis] ? "*" : "",
-                           design_pos[axis] / 65536.0 );
-            strncat( Header, temp,
-                     sizeof ( Header ) - strlen( Header ) - 1 );
-          }
-        }
-        grWriteCellString( bit, 0, 24, Header, fore_color );
-
-        if ( num_shown_axes > MAX_MM_AXES / 2 )
-        {
-          unsigned int  limit = num_shown_axes;
-
-
-          sprintf( Header, "     " );
-
-          for ( n = MAX_MM_AXES / 2; n < limit; n++ )
-          {
-            char  temp[100];
-            int   axis;
-
-
-            axis = shown_axes[n];
-
-            sprintf( temp, "  %.50s%s: %.02f",
-                           multimaster->axis[axis].name,
-                           hidden[axis] ? "*" : "",
-                           design_pos[axis] / 65536.0 );
-            strncat( Header, temp,
-                     sizeof ( Header ) - strlen( Header ) - 1 );
-          }
-
-          grWriteCellString( bit, 0, 32, Header, fore_color );
+          strbuf_reset( header );
+          strbuf_format( header, "%c %.50s%s: %.02f",
+                         n + 'A',
+                         multimaster->axis[axis].name,
+                         hidden[axis] ? "*" : "",
+                         design_pos[axis] / 65536.0 );
+          grWriteCellString( bit, 0, (int)( n + 4 ) * HEADER_HEIGHT,
+                             Header, fore_color );
         }
 
         {
@@ -1230,6 +1208,7 @@
 
           const char*  format_str = NULL;
 
+
           if ( !strcmp( font_format, "CFF" ) )
             format_str = ( cff_hinting_engine == FT_HINTING_FREETYPE
                          ? "CFF (FreeType)"
@@ -1249,23 +1228,27 @@
                                        ? "TrueType (v38)"
                                        : "TrueType (v40)" ) );
 
-          sprintf( Header, "at %d points, first glyph = %d, format = %s",
-                           ptsize,
-                           Num,
-                           format_str );
+          strbuf_reset( header );
+          strbuf_format(
+            header,
+            "size: %dpt, first glyph: %d, format: %s, axis incr.: %.1f%%",
+            ptsize,
+            Num,
+            format_str,
+            increment / 10.0 );
         }
       }
       else
-        sprintf( Header, "%.100s: not an MM font file, or could not be opened",
-                         ft_basename( argv[file] ) );
+        strbuf_format( header,
+                       "%.100s: not an MM font file, or could not be opened",
+                       ft_basename( argv[file] ) );
 
-      grWriteCellString( bit, 0, 8, Header, fore_color );
-      grRefreshSurface( surface );
+      grWriteCellString( bit, 0, HEADER_HEIGHT, Header, fore_color );
 
       if ( !( key = Process_Event() ) )
         goto End;
 
-      if ( key == 'n' )
+      if ( key == grKEY( '.' ) )
       {
         if ( file_loaded >= 1 )
           FT_Done_Face( face );
@@ -1276,18 +1259,18 @@
         goto NewFile;
       }
 
-      if ( key == 'p' )
+      if ( key == grKEY( ',' ) )
       {
         if ( file_loaded >= 1 )
           FT_Done_Face( face );
 
-        if ( file > 1 )
+        if ( file > 0 )
           file--;
 
         goto NewFile;
       }
 
-      if ( key == 'H' )
+      if ( key == grKeyF6 )
       {
         /* enforce reloading */
         if ( file_loaded >= 1 )
